from enum import Enum, EnumMeta


class EnumWithContains(EnumMeta):
    def __contains__(cls, item):  # noqa: N805
        try:
            cls(item)
        except ValueError:
            return False
        return True


class StrEnum(str, Enum, metaclass=EnumWithContains):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()

    def __eq__(self, other: str) -> bool:  # type: ignore
        return self.value == other

    def __hash__(self) -> int:
        return hash(self.value)



import logging
from contextlib import closing
from typing import Dict, List, Optional

from kafka import KafkaAdminClient, KafkaConsumer
from kafka.admin import NewTopic

log = logging.getLogger(__name__)


class KafkaTopicsManager:
    """Helper class to perform actions on kafka servers."""

    def __init__(self, bootstrap_servers: str):
        self._bootstrap_servers: str = bootstrap_servers

    def check_topic(self, topic: str) -> bool:
        return topic in self.list_topics()

    def create_topic(
        self, topic: str, partitions: int = 1, replication: int = 2, config: Optional[Dict] = None
    ) -> None:
        with closing(KafkaAdminClient(bootstrap_servers=self._bootstrap_servers)) as admin_client:
            admin_client.create_topics([NewTopic(topic, partitions, replication, topic_configs=config)])

    def delete_topic(self, topic: str) -> None:
        with closing(KafkaAdminClient(bootstrap_servers=self._bootstrap_servers)) as admin_client:
            admin_client.delete_topics([topic])

    def list_topics(self) -> List[str]:
        with closing(KafkaConsumer(bootstrap_servers=self._bootstrap_servers)) as consumer:
            return list(consumer.topics())



from datetime import datetime


def is_tz_aware(dt: datetime) -> bool:
    """
    Returns whether the datetime is timezone-aware.

    Source: https://stackoverflow.com/a/27596917/148585
    """
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None



import asyncio
import contextlib
import logging
import time
from multiprocessing import Process, Value
from typing import Optional

import pandas as pd
from aiohttp import web
from aiorun import run
from environs import Env

from .async_utils import safe_ensure_future
from .typing_local import Timestamp_ms

env = Env()
env.read_env()  # read .env file, if it exists
LIVENESS_PORT = env.int("LIVENESS_PORT", 8511)

log = logging.getLogger(__name__)


class LivenessClient:
    """Use this class (make a subclass) to override last success timestamp retrieval."""

    def __init__(self):
        self._last_successful_execution: Timestamp_ms = 0

    @property
    def last_success_timestamp(self) -> pd.Timestamp:
        return pd.Timestamp(self._last_successful_execution, unit='ms')


class WebServer:
    """Liveness API webserver to handle kubernetes liveness probes, intended to be run in a separate process."""

    def __init__(self, last_success_value: Value, max_delay: pd.Timedelta) -> None:  # type: ignore
        """
        Initialize a webserver.

        :param last_success_value: inter-process shared object (Value) to hold last success timestamp
        :param max_delay: maximum delay since last success to threat an underlying app as alive
        """
        self._last_success_value: Value = last_success_value  # type: ignore
        self._max_delay = max_delay

        self._app: web.Application = web.Application()
        self._app.router.add_get("/", self.health_check)

    @property
    def web_app(self) -> web.Application:
        return self._app

    @classmethod
    async def run_server(cls, last_success_value: Value, max_delay: pd.Timedelta, port: int) -> None:  # type: ignore
        """Launch a listening server."""
        server = cls(last_success_value, max_delay)
        runner: web.AppRunner = web.AppRunner(server.web_app)
        await runner.setup()
        site = web.TCPSite(runner, "0.0.0.0", port)
        await site.start()
        while True:
            try:
                await asyncio.sleep(86400)
            except asyncio.CancelledError:
                await runner.cleanup()
                break

    @classmethod
    def run_in_loop(cls, *args, **kwargs):
        """Use this when you're starting the server from a separate process."""
        run(cls.run_server(*args, **kwargs), stop_on_unhandled_errors=True)

    async def health_check(self, _: web.Request) -> web.Response:
        """
        Liveness health check endpoint.

        Returns good response (200) when monitored app is healthy, and bad (418) when it's not
        """
        timestamp = self._last_success_value.value  # type: ignore
        last_success: pd.Timestamp = pd.Timestamp(timestamp, unit="s", tz="utc")
        delay: pd.Timedelta = pd.Timestamp.utcnow() - last_success
        headers = {"Last-Success": last_success.isoformat()}
        if delay < self._max_delay:
            return web.Response(text="OK", headers=headers)
        else:
            return web.Response(status=418, text="I'm a teapot", headers=headers)


class LivenessAPIV2:
    """
    Kubernetes liveness check handler, which uses a separate process.

    Rationale: when liveness API is implemented in the same process as monitored app, it could fail to respond in
    time whether the underlying app is really busy doing some CPU-intensive work, thus triggering a container
    restart. Running in separate process allows to monitor a busy app much better.

    The monitored client should inherit from LivenessClient and implement `last_success_timestamp` property (or set
    self._last_successful_execution attribute).

    Intended to be used as a context manager:

    .. code-block:: python

        worker = MyApp()
        liveness_api = LivenessAPIV2(worker)
        async with liveness_api.start():
            worker.start()
    """

    _default_delay = pd.Timedelta(hours=1)

    def __init__(
        self,
        client: LivenessClient,
        max_delay: pd.Timedelta = _default_delay,
        liveness_port: int = LIVENESS_PORT,
    ):
        self._last_success_timestamp_val = Value("d", time.time())
        self._client: LivenessClient = client
        self._max_delay: pd.Timedelta = max_delay
        self._liveness_port: int = liveness_port
        self._child_process: Process = Process(
            target=WebServer.run_in_loop,
            args=(self._last_success_timestamp_val, self._max_delay, self._liveness_port),
            daemon=True,
        )
        self._update_last_success_task: Optional[asyncio.Task] = None

    @property
    def last_success(self) -> pd.Timestamp:
        return pd.Timestamp(self._last_success_timestamp_val.value, unit="s", tz="UTC")

    @property
    def liveness_port(self) -> int:
        return self._liveness_port

    async def update_last_success_loop(self):
        while True:
            try:
                self._last_success_timestamp_val.value = self._client.last_success_timestamp.timestamp()
                await asyncio.sleep(5)
            except asyncio.CancelledError:
                return
            except Exception:
                log.error("Unknown error trying to update last success timestamp.", exc_info=True)
                await asyncio.sleep(30)

    @contextlib.asynccontextmanager
    async def start(self):
        self._child_process.start()
        self._update_last_success_task = safe_ensure_future(self.update_last_success_loop())
        try:
            yield self
        finally:
            self._child_process.terminate()
            if self._update_last_success_task is not None:
                self._update_last_success_task.cancel()



import logging
import threading
from abc import ABC
from typing import Any, Dict, Optional

import boto3
from botocore.client import BaseClient
from botocore.paginate import PageIterator

log = logging.getLogger(__name__)


class AwsBase(ABC):
    _service: str

    def __init__(  # type: ignore
        self,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        region_name: str = "us-west-2",
        **kwargs
    ) -> None:  # type: ignore

        if (access_key is not None) ^ (secret_key is not None):
            raise ValueError("You must specify both `access_key` and `secret_key` or neither of them")

        self._access_key = access_key
        self._secret_key = secret_key
        self._region_name = region_name
        self._kwargs = kwargs
        self._client: Optional[BaseClient] = None
        self._client_creation_lock = threading.Lock()

    @property
    def client(self) -> BaseClient:
        # Unfortunately creation of boto3 client is NOT thread-safe
        with self._client_creation_lock:
            if not self._client:
                self._client = boto3.client(
                    self._service,
                    aws_access_key_id=self._access_key,
                    aws_secret_access_key=self._secret_key,
                    region_name=self._region_name,
                    **self._kwargs
                )
        return self._client

    def _get_page_iterator(self, command: str, operation_params: Dict[str, Any]) -> PageIterator:
        """
        Obtain page iterator for an operation.

        Mostly needed to mock API responses from unittests.
        """
        paginator = self.client.get_paginator(command)
        return paginator.paginate(**operation_params)



import logging
import os
import shutil
import socket
import time
import uuid
from contextlib import contextmanager
from pathlib import Path, PurePath
from typing import Any, Callable, ContextManager, Generator, NamedTuple

import docker
import pytest
import sqlalchemy
from kafka import KafkaConsumer
from kafka.errors import UnrecognizedBrokerVersion
from redis.sentinel import MasterNotFoundError, Sentinel
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import close_all_sessions
from sqlalchemy_utils import create_database, database_exists, drop_database

from birdfeeder.kafka import KafkaTopicsManager

log = logging.getLogger(__name__)


class KafkaCluster(NamedTuple):
    zookeeper: Any
    broker: Any
    bootstrap_servers: str


@pytest.fixture(scope="session")
def session_id() -> str:
    """
    Generate unique session id.

    This is needed in case testsuite may run in parallel on the same server, for example if CI/CD is being used. CI/CD
    infrastructure may run tests for each commit, so these tests should not influence each other.
    """
    return str(uuid.uuid4())


@pytest.fixture(scope="session")
def unused_port():
    """Obtain unused port to bind some service."""

    def _unused_port():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("127.0.0.1", 0))
            return s.getsockname()[1]

    return _unused_port


@pytest.fixture(scope="session")
def docker_manager():
    """Initialize docker management client."""
    return docker.from_env(version="auto")


@pytest.fixture(scope="session")
def kafka_cluster(session_id, unused_port, docker_manager):
    """Run kafka cluster in docker containers."""
    network = docker_manager.networks.create(f"kafka-test-{session_id}")

    zk_port = unused_port()
    zookeeper = docker_manager.containers.run(
        image="wurstmeister/zookeeper:latest",
        name=f"zookeeper-{session_id}",
        network=network.name,
        ports={"2181": zk_port},
        environment={"ZOOKEEPER_CLIENT_PORT": 2181},
        detach=True,
    )

    broker_port = unused_port()
    broker = docker_manager.containers.run(
        image="wurstmeister/kafka:2.12-2.2.2",
        name="broker-{}".format(session_id),
        network=network.name,
        ports={broker_port: broker_port},
        environment={
            "KAFKA_BROKER_ID": 0,
            "KAFKA_ZOOKEEPER_CONNECT": f"zookeeper-{session_id}:2181",
            "KAFKA_ADVERTISED_LISTENERS": f"INSIDE://127.0.0.1:9092,OUTSIDE://127.0.0.1:{broker_port}",
            "KAFKA_LISTENERS": f"INSIDE://:9092,OUTSIDE://:{broker_port}",
            "KAFKA_LISTENER_SECURITY_PROTOCOL_MAP": "INSIDE:PLAINTEXT,OUTSIDE:PLAINTEXT",
            "KAFKA_INTER_BROKER_LISTENER_NAME": "INSIDE",
            "KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR": 1,
            "KAFKA_MESSAGE_MAX_BYTES": 104857600,
        },
        detach=True,
    )
    zookeeper.service_port = zk_port
    broker.service_port = broker_port
    yield KafkaCluster(zookeeper=zookeeper, broker=broker, bootstrap_servers=f"127.0.0.1:{broker.service_port}")

    # Cleanup everything after test run
    zookeeper.remove(v=True, force=True)
    broker.remove(v=True, force=True)
    network.remove()


def wait_kafka(kafka_cluster: KafkaCluster, retries: int = 60) -> None:
    """Waits for kafka cluster to be ready."""
    while retries:
        try:
            with socket.create_connection(("127.0.0.1", kafka_cluster.broker.service_port), timeout=5):
                consumer = KafkaConsumer(bootstrap_servers=kafka_cluster.bootstrap_servers)
                consumer.close()
                break
        except (
            OSError,
            ValueError,
            UnrecognizedBrokerVersion,
        ):
            if not retries:
                raise
            retries -= 1
            log.debug(f"Kafka retries left: {retries}")
            time.sleep(1)


@pytest.fixture()
def kafka(kafka_cluster: KafkaCluster, caplog: pytest.LogCaptureFixture) -> Generator[KafkaCluster, None, None]:
    """Helper fixture to cleanup kafka topic after each test."""
    caplog.set_level(logging.WARNING, "kafka")
    caplog.set_level(logging.WARNING, "aiokafka")
    wait_kafka(kafka_cluster)
    manager = KafkaTopicsManager(kafka_cluster.bootstrap_servers)
    yield kafka_cluster

    topics = manager.list_topics()
    for topic in topics:
        manager.delete_topic(topic)


@pytest.fixture(scope="session")
def mysql(session_id, unused_port, docker_manager):
    """Creates MySQL docker container."""
    port = unused_port()
    password = "pass"
    volume = f"/tmp/mysql-data/{session_id}" if os.getenv("CI") else f"birdfeeder-mysql-test-{session_id}"
    options = " ".join(
        [
            "--innodb-flush-log-at-trx-commit=0",
            "--skip-innodb-doublewrite",
            "--innodb-flush-method=O_DIRECT_NO_FSYNC",
            "--innodb-file-per-table=OFF",
            "--sql-mode=''",
        ]
    )

    container = docker_manager.containers.run(
        image="mysql:5.7",
        # Setting some command-line options to speed up test execution and to bring AWS settings
        command=options,
        name=f"mysql-{session_id}",
        ports={"3306": port},
        environment={"MYSQL_ROOT_PASSWORD": password},
        # Speed up MySQL startup time, approx 30 seconds
        volumes=[f"{volume}:/var/lib/mysql:rw"],
        detach=True,
        # needed for compatibility with Apple Silicon computers
        # becausre there's no ARM version of MySQL 5.7 image
        platform="linux/x86_64",
    )
    container.service_port = port
    container.user = "root"
    container.host = "127.0.0.1"
    container.password = password

    try:
        yield container
    finally:
        # Cleanup everything after test run
        container.remove(v=True, force=True)


def wait_db(url: str, retries: int = 60) -> None:
    """Waits for server readiness."""
    while retries:
        try:
            database_exists(url)
            return
        except sqlalchemy.exc.OperationalError:
            if not retries:
                raise
            retries -= 1
            log.debug(f"MySQL retries left: {retries}")
            time.sleep(1)


@pytest.fixture(scope="session")
def get_new_db() -> Callable[[docker.models.containers.Container, str], ContextManager[URL]]:
    @contextmanager
    def _get_new_db(mysql: docker.models.containers.Container, session_id: str) -> Generator[URL, None, None]:
        db_name = f"parrot-{session_id}"
        db_url = URL.create(
            "mysql+mysqldb",
            username=mysql.user,
            password=mysql.password,
            host=mysql.host,
            port=mysql.service_port,
            database=db_name,
        )

        wait_db(db_url)
        create_database(db_url)
        try:
            yield db_url
        finally:
            # if a test case leave some SQL Sessions open at this point
            # then the database will keep locks for them
            # and "DROP" operation below will freeze current run of test suite
            # so we should close all active sessions manually
            # keep in mind that your tests shouldn't create locks for shared resources
            close_all_sessions()
            drop_database(db_url)

    return _get_new_db


@pytest.fixture(scope="class")
def redis_sentinel_config_template_path():
    """
    Provides python-style template for redis sentinel config, as a path.

    Override this fixture in your project
    """
    here = Path(__file__).parent
    path = PurePath(here).joinpath(".", "sample_configs", "redis_sentinel_config")
    return path


@pytest.fixture(scope="class")
def redis_cluster(session_id, unused_port, docker_manager, redis_sentinel_config_template_path):

    volume = f"/tmp/redis_config_{session_id}"
    Path(volume).mkdir(exist_ok=True)

    network = docker_manager.networks.create(f"redis-test-{session_id}")

    master_host = f"redis-master-{session_id}"
    master_port = unused_port()

    slave_host = f"redis-slave-{session_id}"
    slave_port = unused_port()

    sentinel_host = f"redis-sentinel-{session_id}"
    sentinel_port = unused_port()

    with open(redis_sentinel_config_template_path) as fd:
        rendered_config = fd.read().format(**vars())
        with open(Path(volume) / "sentinel", "w") as target:
            target.write(rendered_config)

    master = docker_manager.containers.run(
        image="redis:6.2.6",
        name=master_host,
        network_mode="host",
        detach=True,
        command=f"redis-server --port {master_port}",
    )

    slave = docker_manager.containers.run(
        image="redis:6.2.6",
        name=slave_host,
        network_mode="host",
        detach=True,
        command=f"redis-server --port {slave_port} --replicaof localhost {master_port}",
    )

    sentinel = docker_manager.containers.run(
        image="redis:6.2.6",
        name=sentinel_host,
        network_mode="host",
        detach=True,
        volumes=[f"{volume}:/redis_configs:rw"],
        command=(
            'bash -c "chown redis:redis /redis_configs/sentinel ' '&& redis-server /redis_configs/sentinel --sentinel"'
        ),
    )

    # wait up to 5 seconds for master to be ready
    try:

        for _ in range(5):
            client = Sentinel([("localhost", sentinel_port)])
            master_conn = client.master_for("mymaster")
            try:
                if master_conn.ping():
                    break
            except MasterNotFoundError:
                time.sleep(1)
                continue
        else:
            pytest.fail("Timeout waiting for initialization of redis master fixture")

        yield "localhost", sentinel_port

    finally:
        master.remove(v=True, force=True)
        slave.remove(v=True, force=True)
        sentinel.remove(v=True, force=True)
        network.remove()

        shutil.rmtree(volume)



import logging
from typing import Any, Dict, List

import botocore.exceptions
import typer

from birdfeeder.aws.base import AwsBase
from birdfeeder.aws.exceptions import BucketUnavailable, MaxKeyUnavailable

log = logging.getLogger(__name__)


class AwsS3(AwsBase):
    """Convenience class to wrap AWS S3 boto3 library"s various methods."""

    _service = "s3"

    def download_with_progressbar(self, bucket: str, key: str, file_path: str):  # type: ignore
        obj_meta = self.client.head_object(Bucket=bucket, Key=key)
        size = obj_meta["ContentLength"]
        with typer.progressbar(length=size) as progress:

            def progress_updater(downloaded_size):
                progress.update(downloaded_size)

            self.client.download_file(bucket, key, file_path, Callback=progress_updater)

    def get_object(self, bucket: str, key: str):  # type: ignore
        """Get object from a bucket and return it's body."""
        response = self.client.get_object(Bucket=bucket, Key=key)
        return response["Body"].read().decode("utf-8")

    def generate_presigned(self, bucket: str, file_name: str) -> Dict[str, Any]:
        result: Dict[str, Any] = self.client.generate_presigned_post(
            Bucket=bucket, Key=file_name, Fields={"acl": "public-read"}, Conditions=[{"acl": "public-read"}]
        )
        return result

    def upload_file(self, file_name: str, bucket: str, object_name: str):  # type: ignore
        self.client.upload_file(file_name, bucket, object_name)

    def download_file(self, bucket: str, key: str, file_path: str, enable_progressbar: bool = False):  # type: ignore
        if enable_progressbar:
            return self.download_with_progressbar(bucket, key, file_path)
        self.client.download_file(bucket, key, file_path)

    def is_bucket_exists(self, bucket: str) -> bool:
        """Check whether a bucket is exists or not (or we don't have access)."""
        try:
            self.client.head_bucket(Bucket=bucket)
        except botocore.exceptions.ClientError:
            # An error occurred (403) when calling the HeadBucket operation: Forbidden
            return False
        # Note: docs states that head_bucket() may raise S3.Client.exceptions.NoSuchBucket, but it cannot be found in
        #  source code currently, and it's unknown how to reproduce it.
        return True

    def is_key_exist(self, bucket: str, key: str) -> bool:
        """Check whether an object with key exists inside the bucket."""
        try:
            self.client.head_object(Bucket=bucket, Key=key)
            return True
        except Exception:
            return False

    def list_objects(self, bucket: str, prefix: str = "") -> List[Any]:
        operation_parameters = {"Bucket": bucket, "Prefix": prefix}
        page_iterator = self._get_page_iterator("list_objects_v2", operation_parameters)
        objects = []  # type: ignore

        for page in page_iterator:
            try:
                contents = page["Contents"]  # noqa: VNE002
                objects = objects + contents
            except KeyError:
                pass

        return objects

    def get_max_key(self, bucket: str, prefix: str = "", check_bucket: bool = True) -> str:
        """
        Get most recent key (filename) in a bucket, matching a path prefix.

        :param bucket: name of the bucket to operate
        :param prefix: path prefix
        :param check_bucket: if True, first check whether the bucket is accessible
        :raises: BucketUnavailable when requested bucket is unavailable
        :raises: MaxKeyUnavailable when there are no files matching a prefix pattern
        """
        if check_bucket and not self.is_bucket_exists(bucket):
            raise BucketUnavailable("Bucket does not exists or is unavailable")

        operation_parameters = {"Bucket": bucket, "Prefix": prefix}
        page_iterator = self._get_page_iterator("list_objects_v2", operation_parameters)
        item_with_max_key = None

        def compare_func(element):
            return element["LastModified"]

        for page in page_iterator:
            try:
                contents = sorted(page["Contents"], key=compare_func)  # noqa: VNE002
            except KeyError:
                # KeyError: 'Contents'
                # raised when requesting nonexistent prefix
                raise MaxKeyUnavailable("Requested prefix does not exists")

            current_page_max_item = contents[-1]
            if item_with_max_key:
                item_with_max_key = max(item_with_max_key, current_page_max_item, key=compare_func)
            else:
                item_with_max_key = current_page_max_item

        return item_with_max_key["Key"]  # type: ignore



import time
from collections import deque
from typing import Callable, Deque, Iterable, NamedTuple, Union

from .typing_local import Timestamp_ms


class TimedMetricItem(NamedTuple):
    """Holds a metric item."""

    timestamp: Union[int, float, Timestamp_ms]
    value: Union[int, float]

    @classmethod
    def create_with_value(cls, value: int) -> "TimedMetricItem":
        """Create new item using current timestamp."""
        return cls(time.time(), value)


def summation(items: Iterable[TimedMetricItem]) -> Union[int, float]:
    """Compute a sum of values."""
    return sum(i.value for i in items)


def average(items: Iterable[TimedMetricItem]) -> float:
    """Compute a mean across values."""
    items = list(items)
    return float(sum(i.value for i in items) / len(items))


class TimedAggregator:
    """
    Helper class to compute aggregated metrics.

    Keeps a timeseries data for a given time interval and calculates aggregated value.
    """

    def __init__(
        self,
        window_size_seconds: float,
        aggregation_func: Callable[[Iterable[TimedMetricItem]], Union[int, float]] = summation,
        time_func: Callable[[], float] = time.time,
    ):
        """
        :param window_size_seconds: aggregation window size
        :param aggregation_func: a function to compute an aggregated value
        :param time_func: a function to get current time
        """
        self._window: Deque[TimedMetricItem] = deque()
        self._window_size_seconds: float = window_size_seconds
        self._aggregation_func: Callable[[Iterable[TimedMetricItem]], Union[int, float]] = aggregation_func
        self._time_func: Callable[[], float] = time_func

    @property
    def aggregated_value(self) -> Union[int, float]:
        """Return aggregated value using aggregation function."""
        self.remove_obsolete_values()
        return self._aggregation_func(self._window)

    @property
    def average_value(self) -> float:
        """Return average value."""
        self.remove_obsolete_values()
        return average(self._window)

    def remove_obsolete_values(self) -> None:
        """Remove values which are outside of time window."""
        now: float = self._time_func()
        threshold = now - self._window_size_seconds
        while len(self._window) > 0 and self._window[0].timestamp < threshold:
            self._window.popleft()

    def add(self, value: Union[int, float] = 1) -> None:
        """Append an item to a timeseries."""
        now: float = self._time_func()
        self.remove_obsolete_values()
        self._window.append(TimedMetricItem(now, value))



import json
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, NamedTuple, Set, Tuple

import pandas as pd


class JSONTemplateField(NamedTuple):
    value_key: str  # original key for the value form the exchange message
    encode_type: type  # value type presented in the json data
    value_type: type  # actual value type
    value_name: str  # name of the ExchangeDataFields or name that describe the key if it's not a standardized key

    @property
    def default_value(self):
        return self.encode_type(self.value_type())


class JSONTemplate:
    def __init__(self, template_field_tuples: Set[Tuple[str, type, type, str]]):
        self._fields: Set[JSONTemplateField] = {JSONTemplateField(*field) for field in template_field_tuples}
        self._name_key_map: Dict[str, Any] = {f.value_name: f.value_key for f in self._fields}

    @property
    def name_key_map(self):
        return self._name_key_map

    @property
    def default_dict(self) -> Dict[str, Any]:
        return self.get_default_template_dict(self._fields)

    @classmethod
    def get_default_template_dict(cls, template_fields: Set[JSONTemplateField]) -> Dict[str, Any]:
        return {str(f.value_key): f.default_value for f in template_fields}


def to_valid_json_dict(dictionary: Any, decimal_to_str: bool = False, enum_to_name: bool = True) -> Any:
    """
    Serialize a structure into JSON-ready form.

    Currently supports of transforming of Decimals and Enums into string form.

    :param dictionary: a structure to serialize, could be a dict, list, set
    :param decimal_to_str: convert Decimal to string if True, and to float if False
    :param enum_to_name: convert Enum attribute to it's name if True, and to value if False
    """
    forward_kwargs = {"decimal_to_str": decimal_to_str, "enum_to_name": enum_to_name}

    if isinstance(dictionary, dict):
        return {key: to_valid_json_dict(value, **forward_kwargs) for key, value in dictionary.items()}
    if isinstance(dictionary, (set, list)):
        return [to_valid_json_dict(value, **forward_kwargs) for value in dictionary]
    if isinstance(dictionary, Decimal):
        return str(dictionary) if decimal_to_str else float(dictionary)
    if isinstance(dictionary, Enum):
        return dictionary.name if enum_to_name else dictionary.value

    return dictionary


def json_encode_default(obj: Any) -> Dict[str, Any]:
    """
    JSON encoder to provide serialization for additional types.

    Example usage:

    .. code-block:: python

        json.dump(obj, fd, indent=4, default=json_encode_default)

    :param obj: object to serialize
    """
    if isinstance(obj, Decimal):
        return {'type(Decimal)': str(obj)}
    elif isinstance(obj, pd.Timestamp):
        return {'type(pd.Timestamp)': str(obj)}
    elif isinstance(obj, datetime):
        return {'type(datetime)': obj.isoformat()}
    else:
        raise TypeError(f"{repr(obj)} is not JSON serializable")


def json_decode_hook(obj: Any) -> Any:
    """
    JSON decode hook to deserialize additional types.

    .. code-block:: python

        json.load(fd, object_hook=json_decode_hook)
    """
    if 'type(Decimal)' in obj:
        return Decimal(obj['type(Decimal)'])
    elif 'type(pd.Timestamp)' in obj:
        return pd.Timestamp(obj['type(pd.Timestamp)'])
    elif 'type(datetime)' in obj:
        return datetime.fromisoformat(obj['type(datetime)'])
    return obj


def dump_to_file_as_json(obj: Any, path: str) -> None:
    """
    Dump python object to a file in JSON.

    Objects that are non JSON-serializable are written as a plain string.

    :param obj: python object
    :param path: output file path
    """
    with open(path, "w") as fd:
        json.dump(obj, fd, indent=4, default=json_encode_default)


def load_json_from_file(path: str) -> Any:
    """Load a JSON object from a file."""
    with open(path, "r") as fd:
        return json.load(fd, object_hook=json_decode_hook)



import os
from typing import TypeVar

from pydantic import BaseSettings

BaseSettingsWithConfig_T = TypeVar("BaseSettingsWithConfig_T", bound="BaseSettingsWithConfig")


class BaseSettingsWithConfig(BaseSettings):
    class Config:
        secrets_dir = os.path.join(os.path.dirname(__file__), "secrets")
        env_file = ".env"

    @classmethod
    def for_env(cls, env: str) -> BaseSettingsWithConfig_T:
        """
        Loads local overrides from an environment-specific file.

        Examples:
        - .env.development
        - .env.staging
        - .env.production
        - .env.local
        """
        return cls(_env_file=f".env.{env.lower()}")



import time
from typing import Union

import pandas as pd

from .typing_local import Timestamp_ms, Timestamp_s


def get_current_timestamp() -> Timestamp_ms:
    """Return current timestamp in ms."""
    return int(time.time() * 1000)


def pd_ts_to_timestamp_ms(ts_input: Union[str, int, float, pd.Timestamp]) -> Timestamp_ms:
    """
    Convert pandas.Timestamp or acceptable `ts_input` into Timestamp_ms.

    See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html for reference
    """
    if not isinstance(ts_input, pd.Timestamp):
        timestamp = pd.Timestamp(ts_input)
    else:
        timestamp = ts_input

    ts: Timestamp_ms = int(timestamp.timestamp() * 1e3)
    return ts


def timestamp_ms_to_str(ts: Timestamp_ms) -> str:
    """Convert millisecond timestamp into human-readable form."""
    return f"{pd.Timestamp(ts, unit='ms')}"


def timestamp_s_to_str(ts: Timestamp_s) -> str:
    """Convert unixtime timestamp into human-readable form."""
    return f"{pd.Timestamp(ts, unit='s')}"


def format_timestamp(timestamp: Union[int, float, str], to_format: str) -> Union[int, float]:
    """
    Convert timestamp from ms/ns/seconds to ms int / seconds int / seconds float form.

    Caveats:
    1. Input ms/ns timestamps are assumed to be fixed-length, which constraints operating range to 2001 - 2262
       (64-bit C long).
    2. Input timestamp cannot be in scientific notation!

    :param timestamp: timestamp in float or int or str
    :param to_format: "s" for second, "ms" for millisecond
    :return: timestamp seconds or milliseconds
    """
    timestamp_as_str: str = str(timestamp)
    milliseconds: int
    seconds: float

    if timestamp_as_str.isnumeric() and len(timestamp_as_str) == 13:  # ms in int
        milliseconds = int(timestamp_as_str)
        seconds = int(timestamp_as_str) / 1000
    elif timestamp_as_str.isnumeric() and len(timestamp_as_str) == 19:  # ns in int
        milliseconds = int(int(timestamp_as_str) / 1e6)
        seconds = int(int(timestamp_as_str) / 1e9)
    elif (
        timestamp_as_str.replace(".", "", 1).isdigit() and len(timestamp_as_str.split(".")[0]) == 10
    ):  # second in float
        milliseconds = int(float(timestamp_as_str) * 1000)
        seconds = float(timestamp_as_str)
    else:
        raise ValueError(f"Invalid timestamp format: {timestamp} (len {len(timestamp_as_str)})")

    if to_format == "ms":
        return milliseconds
    elif to_format == "s":
        return seconds
    else:
        raise ValueError("Invalid param to_format expecting [s, ms]")



HOURS_IN_DAY: int = 24
SECONDS_IN_MINUTE: int = 60
SECONDS_IN_HOUR: int = SECONDS_IN_MINUTE * 60
SECONDS_IN_DAY: int = SECONDS_IN_HOUR * 24
MS_PER_SEC: int = 1000
MS_IN_DAY: int = SECONDS_IN_DAY * MS_PER_SEC
MS_IN_HOUR: int = 60 * 60 * MS_PER_SEC
MS_IN_MINUTE: int = 60 * MS_PER_SEC
HOURS_IN_WEEK: int = int(24 * 7)
HOURS_IN_YEAR: int = int((SECONDS_IN_DAY * 365.25) // (60 * 60))



import logging
from typing import Any, Dict, List

import boto3

from birdfeeder.aws.base import AwsBase

log = logging.getLogger(__name__)


class AwsAutoscaling(AwsBase):
    """Helper class to work with Autoscaling Groups."""

    _service = "autoscaling"

    def get_instances_in_asg(self, autoscaling_group: str) -> List[Dict[str, Any]]:
        """
        Returns EC2 instances inside autoscaling group.

        :param autoscaling_group: AWS autoscaling group name
        """
        asg = self.client.describe_auto_scaling_groups(AutoScalingGroupNames=[autoscaling_group])
        instances = asg["AutoScalingGroups"][0]["Instances"]
        return instances

    def get_asg_instances_ip(self, autoscaling_group: str, return_public_ip: bool = False) -> List[str]:
        """
        Return IPs of the instances in the autoscaling group.

        :param autoscaling_group: AWS autoscaling group name
        :param return_public_ip: if True, return public IP of the instances
        """
        instances = self.get_instances_in_asg(autoscaling_group)
        ec2 = boto3.resource("ec2", region_name=self._region_name)
        instances_resource = [ec2.Instance(i["InstanceId"]) for i in instances]
        attr_name = "public_ip_address" if return_public_ip else "private_ip_address"
        return [getattr(i, attr_name) for i in instances_resource]



import io
import logging
import logging.config
import os
from enum import Enum
from typing import Any, Dict, List, Type, TypeVar

from pythonjsonlogger.jsonlogger import JsonFormatter
from ruamel.yaml import YAML

T = TypeVar("T", bound="Formatter")


class Formatter(Enum):
    """Define logging formatter in a simple way."""

    VERBOSE = "%(asctime)s - %(name)s(%(funcName)s:%(lineno)d) - %(levelname)s: %(message)s"
    JSON = "%(asctime)s %(name)s %(pathname)s %(funcName)s %(lineno)d %(levelname)s %(message)s"

    @classmethod
    def all(cls: Type[T]) -> List[str]:
        """Returns list with all defined formatter names."""
        # Looks like mypy can't correctly handle that for now
        return list(map(lambda i: i.name, cls))  # type: ignore


def configure_logging_formatter(formatter: Formatter = Formatter.JSON) -> None:
    """
    Configure formatter for all existing loggers.

    Note: it sets StreamHandler only
    """
    if formatter is Formatter.JSON:
        formatter_instance = JsonFormatter(formatter.value)
    else:
        formatter_instance = logging.Formatter(formatter.value)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter_instance)

    # We're setting a proper formatter on all existing loggers including these which created at import time
    for name in logging.getLogger().manager.loggerDict:  # type: ignore
        logger = logging.getLogger(name)
        logger.addHandler(handler)
        # Prevent twice messages from child loggers like "aiohttp.access"
        logger.propagate = False


def read_logging_config(file_path: str, **kwargs: Any) -> Dict[str, Any]:
    """Read yaml-formatted config into dict."""
    yaml_parser: YAML = YAML()
    with open(file_path) as fd:
        yml_source: str = fd.read()

        for key, value in kwargs.items():
            yml_source = yml_source.replace(f"${key.upper()}", value)

        # Assume that configs are located in `project_dir/something/`
        project_dir = os.path.abspath(os.path.join(os.path.dirname(file_path), ".."))
        yml_source = yml_source.replace("$PROJECT_DIR", project_dir)

        io_stream: io.StringIO = io.StringIO(yml_source)
        config: Dict = yaml_parser.load(io_stream)

    return config


def init_logging(
    config_path: str,
    load_common_config: bool = True,
    keep_existing: bool = True,
    stdout_formatter: str = "verbose",
    **kwargs: Any,
) -> None:
    """
    Read logging configuration from file and configure loggers.

    :param config_path: path to the logging config file
    :param load_common_config: if True, load common_logging.yml and merge config from `config_path`.
        common_logging.yml should be located in the same directory as `config_path`
    :param keep_existing: if True, existing loggers are kept, but reconfigured to propagate to a new root logger
    :param stdout_formatter: should be one of the formatters defined inside config
    :param kwargs: any additional params, they are transformed to upper-case and used to replace $VARIABLEs in
        logging config
    :return:
    """

    logging_config = {}
    kwargs["stdout_formatter"] = stdout_formatter

    if load_common_config:
        dirname = os.path.dirname(config_path)
        common_logging_path = os.path.join(dirname, "common_logging.yml")
        logging_config = read_logging_config(common_logging_path, **kwargs)

    overrides_config = read_logging_config(config_path, **kwargs)
    overrides_config.setdefault("loggers", {})

    if keep_existing:
        name: str
        logger: logging.Logger
        for name, logger in logging.getLogger().manager.loggerDict.items():  # type: ignore
            if not isinstance(logger, logging.PlaceHolder):
                # We're finding out loggers which were initialized by `getLogger()` call, and preserving their exact
                # level, which is logging.NOTSET by default
                overrides_config["loggers"].setdefault(name, {"level": logger.level})

    logging_config.update(overrides_config)
    logging.config.dictConfig(logging_config)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['birdfeeder', 'birdfeeder.aws', 'birdfeeder.enum', 'birdfeeder.pydantic']

package_data = \
{'': ['*'], 'birdfeeder': ['sample_configs/*']}

install_requires = \
['aiohttp>=3.2,<4.0',
 'aiorun>=2020,<2021',
 'boto3>=1,<2',
 'cachetools>=4,<5',
 'environs>=9,<10',
 'kafka-python>=2,<3',
 'pandas>=1.1,<2.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-json-logger>=2,<3',
 'ruamel.yaml>=0.16,<0.17',
 'typer>=0.3']

entry_points = \
{'console_scripts': ['build_image = birdfeeder.build_image:app']}

setup_kwargs = {
    'name': 'birdfeeder-coinalpha',
    'version': '1.0.3',
    'description': 'Helper library for CoinAlpha projects',
    'long_description': '# birdfeeder\n\nHelper library for CoinAlpha projects\n\n## Usage\n\nThe library is published to pypi as [birdfeeder-coinalpha](https://pypi.org/project/birdfeeder-coinalpha/)\n\n### Installing\n\n```\npip install birdfeeder-coinalpha\n```\n\n```\npoetry add \'birdfeeder-coinalpha@^1\'\n```\n\n\n### How to install old versions\n\nIn pyproject.toml:\n\n```\nbirdfeeder-coinalpha = { git = "https://github.com/coinalpha/birdfeeder.git", branch = "master" }\nbirdfeeder-coinalpha = { git = "https://github.com/coinalpha/birdfeeder.git", rev = "29cdd7229d0d35a989322f5026382400d1332da4" }\nbirdfeeder-coinalpha = { git = "https://github.com/coinalpha/birdfeeder.git", tag = "0.1.0" }\n```\n\npip:\n\n```\ngit+https://github.com/coinalpha/birdfeeder.git@master#egg=birdfeeder\ngit+https://github.com/coinalpha/birdfeeder.git@29cdd7229d0d35a989322f5026382400d1332da4#egg=birdfeeder\ngit+https://github.com/coinalpha/birdfeeder.git@0.1.0#egg=birdfeeder\n```\n\n\n## Development\n\nTo install library for development in conda environment, run\n\n```\n./install\n```\n\nAlternativelly (preferred), you can use poetry env:\n\n```\npoetry install\npoetry shell\n```\n\n## How to add a dependency\n\n1. If you\'re running in conda, you need to install required package first.\n1. Then, see installed version in `conda env export`\n1. Take this version and add into pyproject.toml, into `[tool.poetry.dependencies]` (or dev section, if the package is needed for development only). Specify version (or range) according to [Dependency Specification](https://python-poetry.org/docs/dependency-specification/)\n\nIf using poetry:\n\n1. Run `poetry add <package>`\n\n## Releasing a new version\n\nIdea: we\'re keeping own version and dependencies info in `pyproject.toml`, and then generating `setup.py` so that the library could be installed via tools like pip. We\'re also generating `environment.yml` file because we\'re mostly using conda to manage development environments at CoinAlpha.\n\n1. Change version in `pyproject.toml` and `birdfeeder/__init__.py`\n1. Generate `setup.py`: `dephell deps convert`\n1. Generate `environment.yml`: `poetry2conda --dev pyproject.toml environment.yml`\n1. Commit updates `git add -u && git commit`\n1. Create git tag `x.y.z`: `git tag x.y.z`\n1. Run `git push && git push --tags`\n1. Publish to pypi: `poetry publish --build`\n',
    'author': 'Vladimir Kamarzin',
    'author_email': 'vvk@vvk.pp.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/coinalpha/birdfeeder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import math
import statistics
from decimal import Decimal
from typing import Iterable, Sequence, Union, overload


@overload
def safe_div(numerator: Union[int, float], denominator: Union[int, float]) -> Union[int, float]:
    ...


@overload
def safe_div(numerator: Decimal, denominator: Decimal) -> Decimal:
    ...


def safe_div(numerator, denominator):
    return numerator / denominator if denominator else 0.0


def safe_mean(values: Union[Iterable, Sequence]) -> Union[int, float, Decimal]:
    try:
        return statistics.mean(values)
    except statistics.StatisticsError:
        return 0.0


def round_up(number: float, decimals: int = 0) -> float:
    multiplier = 10**decimals
    return math.ceil(number * multiplier) / multiplier



import os
import subprocess
from datetime import datetime
from typing import List, Optional

import typer

DEFAULT_ORGANIZATION = "coinalpha"

app = typer.Typer()


def get_available_images() -> List[str]:
    """Read files in current directory and return `xxx` part from all Dockerfile.xxx."""
    images = [f.split(".")[1] for f in os.listdir(".") if f.startswith("Dockerfile.")]
    return images


def get_org_from_dockerfile(dockerfile: str) -> str:
    """Return organization name by reading Dockerfile `org` label."""
    with open(dockerfile) as fd:
        for line in fd:
            if line.startswith("LABEL org"):
                return line.split("=")[1].strip().strip('"')

    return DEFAULT_ORGANIZATION  # fallback


def build_image(organization: str, dockerfile: str, image: str, tag: str) -> None:
    """Calls `docker build` to build an image."""
    cmd: str = f"docker build -t {organization}/{image}:{tag} -f {dockerfile} ."
    subprocess.check_call(cmd, shell=True)  # noqa: DUO116


def push_image(organization: str, image: str, image_tag: str) -> None:
    """Calls `docker push` to push an image."""
    subprocess.check_call(f"docker push {organization}/{image}:{image_tag}", shell=True)  # noqa: DUO116


def get_short_rev() -> str:
    """Get current git revision in a short form."""
    short_rev = subprocess.check_output("git rev-parse --short HEAD", shell=True).decode("utf8").strip()  # noqa: DUO116
    return short_rev


def get_default_image_tag() -> str:
    """Construct a docker image tag using date and git short revision."""
    short_rev = get_short_rev()
    image_tag: str = f"{datetime.utcnow().strftime('%Y%m%d')}.git-{short_rev}"
    return image_tag


@app.command()
def main(
    image: Optional[str] = typer.Argument(None),  # noqa: B008
    push: bool = typer.Option(False, is_flag=True, help="Push the new image to Docker Hub."),  # noqa: B008
    tag: str = typer.Option(  # noqa: B008
        get_default_image_tag(),  # noqa: B008
        help="Specify a tag to be used, instead of the default one. "
        "Default is to construct a tag using current git revision (short).",
    ),
) -> None:
    """
    Build docker IMAGE.

    Expects to see Dockerfile.IMAGE in the current working directory.
    """
    available_images = get_available_images()
    if image not in available_images:
        typer.echo("Bad or missing image argument.\nTry --help' for help.\n\nAvailable images:")
        for i in available_images:
            typer.echo(f"- {i}")
        raise typer.Exit(code=1)

    dockerfile = f"Dockerfile.{image}"
    organization = get_org_from_dockerfile(dockerfile)
    build_image(organization, dockerfile, image, tag)

    if push:
        push_image(organization, image, tag)


if __name__ == '__main__':
    app()



import asyncio
import functools
import inspect
import logging
import time
from typing import Any, List, Tuple, Union

import cachetools
from async_timeout import timeout
from environs import Env

log = logging.getLogger(__name__)
env = Env()
env.read_env()  # read .env file, if it exists
SHOULD_INSPECT = env.bool("INSPECT_CALLERS", False)


def async_ttl_cache(ttl: int = 3600, maxsize: int = 1) -> Any:
    """Decorator to cache a coroutine result using :py:class:`cachetools.TTLCache`."""
    cache: cachetools.TTLCache = cachetools.TTLCache(ttl=ttl, maxsize=maxsize)

    def decorator(fn):
        @functools.wraps(fn)
        async def memoize(*args, **kwargs):
            key = str((args, kwargs))
            try:
                return cache[key]
            except KeyError:
                result = await fn(*args, **kwargs)
                cache[key] = result
                return result

        return memoize

    return decorator


def get_callers(stack_size: int = 5) -> List[Tuple[int, str, Any]]:
    stack = inspect.stack()
    modules = [(index, inspect.getmodule(stack[index][0])) for index in reversed(range(1, min(stack_size, len(stack))))]
    callers = []
    for index, module in modules:
        try:
            callers.append((index, module.__name__, stack[index][3]))  # type: ignore
        except AttributeError:
            # AttributeError: 'NoneType' object has no attribute '__name__'
            callers.append((index, "unknown", stack[index][3]))
    return sorted(callers)


def safe_ensure_future(coro, *args, **kwargs):
    """
    Run a coroutine in a wrapper, catching and logging unexpected exception.

    :envvar: INSPECT_CALLERS: if true, show callers on failure
    """
    caller_names = ""
    if SHOULD_INSPECT:
        caller_names = "\n" + "\n".join([str(t) for t in get_callers()])

    async def safe_wrapper(c):
        try:
            return await c
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.getLogger().error(
                f"Unhandled error in background task: {str(e)} {caller_names}",
                exc_info=True,
            )

    wrapped_coro = safe_wrapper(coro)

    if coro.__name__ and isinstance(coro.__name__, str) and coro.__name__.isidentifier():
        wrapped_coro.__qualname__ = f"safe_{coro.__qualname__}"

    return asyncio.ensure_future(wrapped_coro, *args, **kwargs)


async def safe_gather(*args, **kwargs):
    """
    Gather an awaitables logging unexpected exceptions.

    .. note::

        Exception is logged and re-raised!

    :envvar: INSPECT_CALLERS: if true, show callers on failure
    """
    caller_names = ""
    if SHOULD_INSPECT:
        caller_names = "\n" + "\n".join([str(t) for t in get_callers()])
    try:
        return await asyncio.gather(*args, **kwargs)
    except Exception as e:
        logging.getLogger().debug(
            f"Unhandled error in background task: {str(e)} {caller_names}",
            exc_info=True,
        )
        raise


async def safe_cancel(task: asyncio.Task, info: str, cancel_wait_timeout: Union[int, float] = 10) -> int:
    """
    Cancel asyncio task and retry cancel if task didn't finished.

    :param task: Task to cancel
    :param info: some additional info to log when cancelling a task
    :param cancel_wait_timeout: how many seconds wait for task to cancel before retrying
    :return: how many attempts has been made to cancel a task
    """
    retries = 0

    while True:
        task.cancel()
        log.debug(f"Task cancel request sent to task: {task}, additional info: {info}")

        try:
            async with timeout(cancel_wait_timeout):
                while not task.done():
                    await asyncio.sleep(cancel_wait_timeout / 10)
        except asyncio.TimeoutError:
            log.warning(f"Task failed to stop: {task}, additional task info: {info}, retries: {retries}")
        else:
            log.info(f"Task cancelled: {task}, additional info: {info}")
            return retries

        retries += 1


def calc_delay_til_next_tick(seconds: float) -> float:
    """Calculate the delay to next tick."""
    now: float = time.time()
    current_tick: int = int(now // seconds)
    delay_til_next_tick: float = (current_tick + 1) * seconds - now
    return delay_til_next_tick


async def wait_til_next_tick(seconds: float = 1.0) -> None:
    """Wait until the end of quantized time interval."""
    delay = calc_delay_til_next_tick(seconds)
    await asyncio.sleep(delay)
